from typing import BinaryIO

from kloppy.config import get_config
from kloppy.exceptions import AdapterError, InputNotFoundError
from .adapter import Adapter

try:
    from js import XMLHttpRequest

    RUNS_IN_BROWSER = True
except ImportError:
    RUNS_IN_BROWSER = False


def check_requests_patch():
    if RUNS_IN_BROWSER:
        try:
            import pyodide_http
        except ImportError:
            raise AdapterError(
                "Seems like you don't have `pyodide-http` installed, which is required to make http requests "
                "work in the browser. Please install it using: pip install pyodide-http"
            )

        pyodide_http.patch_all()


class HTTPAdapter(Adapter):
    def supports(self, url: str) -> bool:
        return url.startswith("http://") or url.startswith("https://")

    def read_to_stream(self, url: str, output: BinaryIO):
        check_requests_patch()

        basic_authentication = get_config("adapters.http.basic_authentication")

        try:
            import requests
        except ImportError:
            raise AdapterError(
                "Seems like you don't have `requests` installed. Please"
                " install it using: pip install requests"
            )

        auth = None
        if basic_authentication:
            auth = requests.auth.HTTPBasicAuth(*basic_authentication)

        with requests.get(url, stream=True, auth=auth) as r:
            if r.status_code == 404:
                raise InputNotFoundError(f"Could not find {url}")

            r.raise_for_status()
            for chunk in r.iter_content(chunk_size=8192):
                output.write(chunk)



from dataclasses import dataclass

from kloppy.utils import docstring_inherit_attributes
from ..event import PassEvent, ShotEvent
from ... import EventFactory, create_event


@dataclass(repr=False)
@docstring_inherit_attributes(PassEvent)
class StatsBombPassEvent(PassEvent):
    """
    Attributes:

    """


@dataclass(repr=False)
@docstring_inherit_attributes(ShotEvent)
class StatsBombShotEvent(ShotEvent):
    """
    Attributes:

    """


class StatsBombEventFactory(EventFactory):
    def build_pass(self, **kwargs) -> StatsBombPassEvent:
        return create_event(StatsBombPassEvent, **kwargs)

    def build_shot(self, **kwargs) -> StatsBombShotEvent:
        return create_event(StatsBombShotEvent, **kwargs)



import os
from contextlib import contextmanager
from copy import copy
from typing import Any, Optional

from kloppy.domain import EventFactory

try:
    from typing import TypedDict
except ImportError:
    from mypy_extensions import TypedDict

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

cache_dir = os.environ.get("KLOPPY_CACHE_DIR")
if not cache_dir:
    cache_dir = os.path.expanduser("~/kloppy_cache")

Config = TypedDict(
    "Config",
    {
        "cache": Optional[str],
        "coordinate_system": Optional[str],
        "event_factory": Optional[EventFactory],
        "adapters.http.basic_authentication": Optional[str],
        "adapters.s3.s3fs": Optional[Any],
    },
)

# https://github.com/python/mypy/issues/6262
CONFIG_KEYS = Literal[
    "cache",
    "coordinate_system",
    "event_factory",
    "adapters.http.basic_authentication",
    "adapters.s3.s3fs",
]


class PartialConfig(Config, total=False):
    pass


_default_config: Config = {
    "cache": cache_dir,
    "coordinate_system": "kloppy",
    "event_factory": None,
    "adapters.http.basic_authentication": None,
    "adapters.s3.s3fs": None,
}

config = copy(_default_config)


def reset_config():
    for key, value in _default_config.items():
        set_config(key, value)  # type: ignore


def set_config(key: CONFIG_KEYS, value: Optional[str]):
    if key in config:
        config[key] = value
    else:
        raise KeyError(f"Non existing config '{key}'")


def get_config(key: Optional[CONFIG_KEYS] = None):
    if key is None:
        return config
    elif key in config:
        return config[key]  # type: ignore
    else:
        raise KeyError(f"Non existing config '{key}'")


@contextmanager
def config_context(*args):
    """Set some config items for within a certain context. Code borrowed partly from
    pandas."""
    if len(args) % 2 != 0 or len(args) < 2:
        raise ValueError(
            "Need to invoke as config_context(key, value, [(key, value), ...])."
        )

    configs = list(zip(args[::2], args[1::2]))

    undo = {key: get_config(key) for key, _ in configs}
    try:
        for key, value in configs:
            set_config(key, value)

        yield

    finally:
        for key, value in undo.items():
            set_config(key, value)



# detect if we are imported from the setup procedure (borrowed from numpy code)
try:
    __KLOPPY_SETUP__
except NameError:
    __KLOPPY_SETUP__ = False

# if not __KLOPPY_SETUP__:
#     from .infra.serializers import *
#     from .helpers import *
#    from .infra import datasets
#     from .domain.services.matchers.pattern import (
#         event as event_pattern_matching,
#     )
#     from .domain.services.state_builder import add_state

__version__ = "3.7.1"



import dataclasses
import warnings
from dataclasses import fields
from typing import TypeVar, Type

from kloppy.domain import (
    PassEvent,
    ShotEvent,
    GenericEvent,
    TakeOnEvent,
    RecoveryEvent,
    CarryEvent,
    FormationChangeEvent,
    BallOutEvent,
    PlayerOnEvent,
    PlayerOffEvent,
    FoulCommittedEvent,
    CardEvent,
    SubstitutionEvent,
)

T = TypeVar("T")


def create_event(event_cls: Type[T], **kwargs) -> T:
    """
    Do the actual construction of an event.

    This method does a couple of things:
    1. Fill in some arguments when not passed
    2. Pass only arguments that are accepted by the `event_cls`.
       This is required in cases an EventFactory initialized less rich
       Events than data is passed for. E.g. `expected_goal` is passed
       to a regular `ShotEvent`.
       Normally this would break because of an 'Unexpected argument' exception,
       but we filter those arguments out.
    """
    extra_kwargs = {"state": {}}
    if "related_event_ids" not in kwargs:
        extra_kwargs["related_event_ids"] = []

    if "freeze_frame" not in kwargs:
        kwargs["freeze_frame"] = None

    all_kwargs = dict(**kwargs, **extra_kwargs)

    relevant_kwargs = {
        field.name: all_kwargs.get(field.name, field.default)
        for field in fields(event_cls)
        if field.init
        and not (
            field.default == dataclasses.MISSING
            and field.name not in all_kwargs
        )
    }

    if len(relevant_kwargs) < len(all_kwargs):
        skipped_kwargs = set(all_kwargs.keys()) - set(relevant_kwargs.keys())
        warnings.warn(
            f"The following arguments were skipped: {skipped_kwargs}"
        )

    return event_cls(**relevant_kwargs)


class EventFactory:
    def build_pass(self, **kwargs) -> PassEvent:
        return create_event(PassEvent, **kwargs)

    def build_shot(self, **kwargs) -> ShotEvent:
        return create_event(ShotEvent, **kwargs)

    def build_generic(self, **kwargs) -> GenericEvent:
        return create_event(GenericEvent, **kwargs)

    def build_recovery(self, **kwargs) -> RecoveryEvent:
        return create_event(RecoveryEvent, **kwargs)

    def build_take_on(self, **kwargs) -> TakeOnEvent:
        return create_event(TakeOnEvent, **kwargs)

    def build_carry(self, **kwargs) -> CarryEvent:
        return create_event(CarryEvent, **kwargs)

    def build_formation_change(self, **kwargs) -> FormationChangeEvent:
        return create_event(FormationChangeEvent, **kwargs)

    def build_ball_out(self, **kwargs) -> BallOutEvent:
        return create_event(BallOutEvent, **kwargs)

    def build_player_on(self, **kwargs) -> PlayerOnEvent:
        return create_event(PlayerOnEvent, **kwargs)

    def build_player_off(self, **kwargs) -> PlayerOffEvent:
        return create_event(PlayerOffEvent, **kwargs)

    def build_card(self, **kwargs) -> CardEvent:
        return create_event(CardEvent, **kwargs)

    def build_foul_committed(self, **kwargs) -> FoulCommittedEvent:
        return create_event(FoulCommittedEvent, **kwargs)

    def build_substitution(self, **kwargs) -> SubstitutionEvent:
        return create_event(SubstitutionEvent, **kwargs)



import os
from typing import BinaryIO

import pytest
from kloppy import opta
from kloppy.config import config_context
from kloppy.domain import Provider, DatasetType
from kloppy.exceptions import InputNotFoundError
from kloppy.infra.io.adapters import Adapter, adapters


class TestAdapter:
    @pytest.fixture
    def f24_data(self) -> str:
        base_dir = os.path.dirname(__file__)
        return f"{base_dir}/files/opta_f24.xml"

    @pytest.fixture
    def f7_data(self) -> str:
        base_dir = os.path.dirname(__file__)
        return f"{base_dir}/files/opta_f7.xml"

    def test_custom_adapter(self, f24_data: str, f7_data: str):
        """
        Make sure the coordinate system can be set and is used
        when loading data.
        """

        class CustomAdapter(Adapter):
            def supports(self, url: str) -> bool:
                return url.startswith("test123://")

            def read_to_stream(self, url: str, output: BinaryIO):
                if url == "test123://f24":
                    fp = open(f24_data, "rb")
                elif url == "test123://f7":
                    fp = open(f7_data, "rb")
                else:
                    raise Exception(f"Unknown url {url}")

                output.write(fp.read())
                fp.close()

        with config_context("cache", None):
            with pytest.raises(InputNotFoundError):
                # When our adapter is not registered yet
                # kloppy will fall back to regular `open`.
                opta.load(f24_data="test123://f24", f7_data="test123://f7")

            custom_adapter = CustomAdapter()
            adapters.append(custom_adapter)

            dataset = opta.load(
                f24_data="test123://f24", f7_data="test123://f7"
            )

            # cleanup
            adapters.remove(custom_adapter)

            # Asserts borrowed from `test_opta.py`
            assert dataset.metadata.provider == Provider.OPTA
            assert dataset.dataset_type == DatasetType.EVENT
            assert len(dataset.events) == 21



import os

import pytest
from kloppy import statsbomb

from kloppy.domain import EventDataset, Point
from kloppy.domain.services.transformers.attribute import (
    DistanceToGoalTransformer,
    DistanceToOwnGoalTransformer,
)


class TestToRecords:
    @pytest.fixture
    def event_data(self) -> str:
        base_dir = os.path.dirname(__file__)
        return f"{base_dir}/files/statsbomb_event.json"

    @pytest.fixture
    def lineup_data(self) -> str:
        base_dir = os.path.dirname(__file__)
        return f"{base_dir}/files/statsbomb_lineup.json"

    @pytest.fixture
    def dataset(self, event_data: str, lineup_data: str) -> EventDataset:
        return statsbomb.load(
            lineup_data=lineup_data,
            event_data=event_data,
            coordinates="statsbomb",
        )

    def test_default_columns(self, dataset: EventDataset):
        records = dataset.to_records()
        assert len(records) == 4023
        assert list(records[0].keys()) == [
            "event_id",
            "event_type",
            "result",
            "success",
            "period_id",
            "timestamp",
            "end_timestamp",
            "ball_state",
            "ball_owning_team",
            "team_id",
            "player_id",
            "coordinates_x",
            "coordinates_y",
        ]

    def test_string_columns(self, dataset: EventDataset):
        """
        When string columns passed to to_records the data must be searched from:
        1. the output of Default()
        2. attributes of the event
        """

        records = dataset.filter("pass").to_records(
            "timestamp", "coordinates_x", "coordinates"
        )
        assert records[0] == {
            "timestamp": 0.098,
            "coordinates_x": 60.5,
            "coordinates": Point(x=60.5, y=40.5),
        }

    def test_string_wildcard_columns(self, dataset: EventDataset):
        """
        Make sure it's possible to specify wildcard pattern to match attributes.
        """

        records = dataset.filter("pass").to_records(
            "timestamp",
            "player_id",
            "coordinates_*",
            DistanceToGoalTransformer(),
            DistanceToOwnGoalTransformer(),
        )
        assert records[0] == {
            "timestamp": 0.098,
            "player_id": "6581",
            "coordinates_x": 60.5,
            "coordinates_y": 40.5,
            "distance_to_goal": 59.50210080324896,
            "distance_to_own_goal": 60.502066080424065,
        }



from typing import Optional

from .adapter import Adapter
from .http import HTTPAdapter
from .s3 import S3Adapter

adapters = [HTTPAdapter(), S3Adapter()]


def get_adapter(url: str) -> Optional[Adapter]:
    for adapter in adapters:
        if adapter.supports(url):
            return adapter



import os

import pytest

from kloppy.domain import (
    AttackingDirection,
    Ground,
    Orientation,
    Period,
    Point,
    Provider,
    SetPieceType,
    DatasetType,
)

from kloppy import datafactory


class TestDatafactory:
    @pytest.fixture
    def event_data(self) -> str:
        base_dir = os.path.dirname(__file__)
        return f"{base_dir}/files/datafactory_events.json"

    def test_correct_deserialization(self, event_data: str):
        dataset = datafactory.load(
            event_data=event_data, coordinates="datafactory"
        )

        assert dataset.metadata.provider == Provider.DATAFACTORY
        assert dataset.dataset_type == DatasetType.EVENT
        assert len(dataset.events) == 1027
        assert len(dataset.metadata.periods) == 2
        assert dataset.events[10].ball_owning_team == dataset.metadata.teams[1]
        assert dataset.events[23].ball_owning_team == dataset.metadata.teams[0]
        assert dataset.metadata.orientation == Orientation.HOME_TEAM
        assert dataset.metadata.teams[0].name == "Team A"
        assert dataset.metadata.teams[0].ground == Ground.HOME
        assert dataset.metadata.teams[1].name == "Team B"
        assert dataset.metadata.teams[1].ground == Ground.AWAY

        player = dataset.metadata.teams[0].players[0]
        assert player.player_id == "38804"
        assert player.jersey_no == 1
        assert str(player) == "Daniel Bold"
        assert player.position is None  # not set
        assert player.starting

        assert dataset.metadata.periods[0] == Period(
            id=1,
            start_timestamp=0,
            end_timestamp=2912,
            attacking_direction=AttackingDirection.HOME_AWAY,
        )
        assert dataset.metadata.periods[1] == Period(
            id=2,
            start_timestamp=2700,
            end_timestamp=5710,
            attacking_direction=AttackingDirection.AWAY_HOME,
        )

        assert dataset.events[0].coordinates == Point(0.01, 0.01)

        # Check the qualifiers
        assert dataset.events[0].qualifiers[0].value == SetPieceType.KICK_OFF
        assert dataset.events[412].qualifiers[0].value == SetPieceType.THROW_IN

    def test_correct_normalized_deserialization(self, event_data: str):
        dataset = datafactory.load(event_data=event_data)

        assert dataset.events[0].coordinates == Point(0.505, 0.505)



from typing import Optional, List

from kloppy.config import get_config
from kloppy.domain import EventDataset, EventFactory
from kloppy.infra.serializers.event.sportec import (
    SportecEventDeserializer,
    SportecInputs,
)
from kloppy.io import open_as_file


def load(
    event_data: str,
    meta_data: str,
    event_types: Optional[List[str]] = None,
    coordinates: Optional[str] = None,
    event_factory: Optional[EventFactory] = None,
) -> EventDataset:
    """
    Load Sportec event data into a [`EventDataset`][kloppy.domain.models.event.EventDataset]

    Parameters:
        event_data: filename of the XML file containing the events
        meta_data: filename of the XML file containing the match information
        event_types:
        coordinates:
        event_factory:

    """
    serializer = SportecEventDeserializer(
        event_types=event_types,
        coordinate_system=coordinates,
        event_factory=event_factory or get_config("event_factory"),
    )
    with open_as_file(event_data) as event_data_fp, open_as_file(
        meta_data
    ) as meta_data_fp:
        return serializer.deserialize(
            SportecInputs(event_data=event_data_fp, meta_data=meta_data_fp)
        )



from typing import Optional
import contextlib

from kloppy.domain import TrackingDataset
from kloppy.infra.serializers.tracking.secondspectrum import (
    SecondSpectrumDeserializer,
    SecondSpectrumInputs,
)
from kloppy.io import FileLike, open_as_file, Source


def load(
    meta_data: FileLike,
    raw_data: FileLike,
    additional_meta_data: Optional[FileLike] = None,
    sample_rate: Optional[float] = None,
    limit: Optional[int] = None,
    coordinates: Optional[str] = None,
    only_alive: Optional[bool] = False,
) -> TrackingDataset:
    deserializer = SecondSpectrumDeserializer(
        sample_rate=sample_rate,
        limit=limit,
        coordinate_system=coordinates,
        only_alive=only_alive,
    )
    with open_as_file(meta_data) as meta_data_fp, open_as_file(
        raw_data
    ) as raw_data_fp, open_as_file(
        Source.create(additional_meta_data, optional=True)
    ) as additional_meta_data_fp:
        return deserializer.deserialize(
            inputs=SecondSpectrumInputs(
                meta_data=meta_data_fp,
                raw_data=raw_data_fp,
                additional_meta_data=additional_meta_data_fp,
            )
        )



from typing import Union

from kloppy.config import get_config
from kloppy.infra.serializers.event.wyscout import (
    WyscoutDeserializer,
    WyscoutInputs,
)
from kloppy.domain import EventDataset, Optional, List, EventFactory
from kloppy.io import open_as_file, FileLike


def load(
    event_data: FileLike,
    event_types: Optional[List[str]] = None,
    coordinates: Optional[str] = None,
    event_factory: Optional[EventFactory] = None,
) -> EventDataset:
    """
    Load Wyscout event data into a [`EventDataset`][kloppy.domain.models.event.EventDataset]

    Parameters:
        event_data: filename of the XML file containing the events and metadata
        event_types:
        coordinates:
        event_factory:
    """
    deserializer = WyscoutDeserializer(
        event_types=event_types,
        coordinate_system=coordinates,
        event_factory=event_factory or get_config("event_factory"),
    )
    with open_as_file(event_data) as event_data_fp:
        return deserializer.deserialize(
            inputs=WyscoutInputs(event_data=event_data_fp),
        )


def load_open_data(
    match_id: Union[str, int] = "2499841",
    event_types: Optional[List[str]] = None,
    coordinates: Optional[str] = None,
    event_factory: Optional[EventFactory] = None,
) -> EventDataset:
    return load(
        event_data=f"https://raw.githubusercontent.com/koenvo/wyscout-soccer-match-event-dataset/main/processed/files/{match_id}.json",
        event_types=event_types,
        coordinates=coordinates,
        event_factory=event_factory,
    )



import contextlib
import warnings
from typing import Union

from kloppy.config import get_config
from kloppy.domain.models.statsbomb.event import StatsBombEventFactory
from kloppy.infra.serializers.event.statsbomb import (
    StatsBombDeserializer,
    StatsBombInputs,
)
from kloppy.domain import EventDataset, Optional, List, EventFactory
from kloppy.io import open_as_file, FileLike, Source


@contextlib.contextmanager
def dummy_context_mgr():
    yield None


def load(
    event_data: FileLike,
    lineup_data: FileLike,
    three_sixty_data: Optional[FileLike] = None,
    event_types: Optional[List[str]] = None,
    coordinates: Optional[str] = None,
    event_factory: Optional[EventFactory] = None,
) -> EventDataset:
    """
    Load StatsBomb event data into a [`EventDataset`][kloppy.domain.models.event.EventDataset]

    Parameters:
        event_data: filename of json containing the events
        lineup_data: filename of json containing the lineup information
        three_sixty_data: filename of json containing the 360 data
        event_types:
        coordinates:
        event_factory:
    """
    deserializer = StatsBombDeserializer(
        event_types=event_types,
        coordinate_system=coordinates,
        event_factory=event_factory
        or get_config("event_factory")
        or StatsBombEventFactory(),
    )
    with open_as_file(event_data) as event_data_fp, open_as_file(
        lineup_data
    ) as lineup_data_fp, open_as_file(
        Source.create(three_sixty_data, optional=True)
    ) as three_sixty_data_fp:

        return deserializer.deserialize(
            inputs=StatsBombInputs(
                event_data=event_data_fp,
                lineup_data=lineup_data_fp,
                three_sixty_data=three_sixty_data_fp,
            ),
        )


def load_open_data(
    match_id: Union[str, int] = "15946",
    event_types: Optional[List[str]] = None,
    coordinates: Optional[str] = None,
    event_factory: Optional[EventFactory] = None,
) -> EventDataset:
    warnings.warn(
        "\n\nYou are about to use StatsBomb public data."
        "\nBy using this data, you are agreeing to the user agreement. "
        "\nThe user agreement can be found here: https://github.com/statsbomb/open-data/blob/master/LICENSE.pdf"
        "\n"
    )

    return load(
        event_data=f"https://raw.githubusercontent.com/statsbomb/open-data/master/data/events/{match_id}.json",
        lineup_data=f"https://raw.githubusercontent.com/statsbomb/open-data/master/data/lineups/{match_id}.json",
        three_sixty_data=Source(
            f"https://raw.githubusercontent.com/statsbomb/open-data/master/data/three-sixty/{match_id}.json",
            skip_if_missing=True,
        ),
        event_types=event_types,
        coordinates=coordinates,
        event_factory=event_factory,
    )



import contextlib
import logging
import os
import urllib.parse
from dataclasses import dataclass, replace
from pathlib import PurePath
from typing import Union, IO, BinaryIO, Tuple

from io import BytesIO

from kloppy.config import get_config
from kloppy.exceptions import InputNotFoundError
from kloppy.infra.io.adapters import get_adapter


logger = logging.getLogger(__name__)

_open = open


@dataclass(frozen=True)
class Source:
    data: "FileLike"
    optional: bool = False
    skip_if_missing: bool = False

    @classmethod
    def create(cls, input_: "FileLike", **kwargs):
        if isinstance(input_, Source):
            return replace(input_, **kwargs)
        return Source(data=input_, **kwargs)


FileLike = Union[str, PurePath, bytes, IO[bytes], Source]


def get_local_cache_stream(url: str, cache_dir: str) -> Tuple[BinaryIO, bool]:
    if not os.path.exists(cache_dir):
        os.makedirs(cache_dir)

    filename = urllib.parse.quote_plus(url)
    local_filename = f"{cache_dir}/{filename}"

    # Open the file in append+read mode
    # this makes sure:
    # 1. The file is created when it does not exist
    # 2. The file is not truncated when it does exist
    # 3. The file can be read
    return _open(local_filename, "a+b"), (
        os.path.exists(local_filename)
        and os.path.getsize(local_filename) > 0
        and local_filename
    )


@contextlib.contextmanager
def dummy_context_mgr():
    yield None


def open_as_file(input_: FileLike) -> IO:
    if isinstance(input_, Source):
        if input_.data is None and input_.optional:
            # This saves us some additional code in every vendor specific code
            return dummy_context_mgr()

        try:
            return open_as_file(input_.data)
        except InputNotFoundError:
            if input_.skip_if_missing:
                logging.info(f"Input {input_.data} not found. Skipping")
                return dummy_context_mgr()
            raise
    elif isinstance(input_, str) or isinstance(input_, PurePath):
        if isinstance(input_, PurePath):
            input_ = str(input_)
            is_path = True
        else:
            is_path = False

        if not is_path and ("{" in input_ or "<" in input_):
            return BytesIO(input_.encode("utf8"))
        else:
            adapter = get_adapter(input_)
            if adapter:
                cache_dir = get_config("cache")
                if cache_dir:
                    stream, local_cache_file = get_local_cache_stream(
                        input_, cache_dir
                    )
                else:
                    stream = BytesIO()
                    local_cache_file = None

                if not local_cache_file:
                    logger.info(f"Retrieving {input_}")
                    adapter.read_to_stream(input_, stream)
                    logger.info("Retrieval complete")
                else:
                    logger.info(f"Using local cached file {local_cache_file}")
                stream.seek(0)
            else:
                if not os.path.exists(input_):
                    raise InputNotFoundError(f"File {input_} does not exist")

                stream = _open(input_, "rb")
            return stream
    elif isinstance(input_, bytes):
        return BytesIO(input_)
    else:
        return input_



import os

import pytest


from kloppy import statsbomb
from kloppy.domain import EventDataset


class TestEvent:
    """"""

    @pytest.fixture
    def event_data(self) -> str:
        base_dir = os.path.dirname(__file__)
        return f"{base_dir}/files/statsbomb_event.json"

    @pytest.fixture
    def lineup_data(self) -> str:
        base_dir = os.path.dirname(__file__)
        return f"{base_dir}/files/statsbomb_lineup.json"

    @pytest.fixture()
    def dataset(self, lineup_data: str, event_data: str) -> EventDataset:
        return statsbomb.load(
            lineup_data=lineup_data,
            event_data=event_data,
            coordinates="statsbomb",
        )

    def test_navigation(self, dataset: EventDataset):
        """
        Test navigating (next/prev) through events
        """
        passes = dataset.find_all("pass")
        assert passes[0].next("pass") == passes[1]
        assert passes[1].prev("pass") == passes[0]
        assert passes[0].next() == dataset.get_event_by_id(
            "61da36dc-d862-416c-8ee3-1a0cd24dc086"
        )

        goals = dataset.find_all("shot.goal")
        assert len(goals) == 3
        assert goals[0].prev("shot.goal") is None
        assert goals[0].next("shot.goal") == goals[1]
        assert goals[0].next("shot.goal") == goals[2].prev("shot.goal")
        assert goals[2].next("shot.goal") is None

        first_goal = dataset.find("shot.goal")
        assert first_goal == goals[0]
        assert first_goal.next(".goal") == goals[1]
        assert first_goal.next(".goal").next(".goal") == goals[2]
        assert first_goal.next(".goal").next(".goal").next(".goal") is None

    def test_filter(self, dataset: EventDataset):
        """
        Test filtering allows simple 'css selector' (<event_type>.<result>)
        """
        goals_dataset = dataset.filter("shot.goal")

        df = goals_dataset.to_pandas()
        assert df["event_id"].to_list() == [
            "4c7c4ab1-6b9f-4504-a237-249c2e0c549f",
            "683c6752-13bc-4892-94ed-22e1c938f1f7",
            "55d71847-9511-4417-aea9-6f415e279011",
        ]

    def test_find_all(self, dataset: EventDataset):
        """
        Test find_all allows simple 'css selector' (<event_type>.<result>)
        """
        goals = dataset.find_all("shot.goal")
        assert len(goals) == 3
        assert goals[0].prev("shot.goal") is None
        assert goals[0].next("shot.goal") == goals[1]
        assert goals[0].next("shot.goal") == goals[2].prev("shot.goal")
        assert goals[2].next("shot.goal") is None



from dataclasses import asdict, fields, replace
from kloppy.domain.models.tracking import PlayerData
from typing import TypeVar, Union

from kloppy.domain import (
    AttackingDirection,
    Dataset,
    DatasetFlag,
    EventDataset,
    Frame,
    Metadata,
    Orientation,
    PitchDimensions,
    Point,
    Point3D,
    Team,
    TrackingDataset,
    CoordinateSystem,
    Origin,
)
from kloppy.domain.models.event import Event
from kloppy.exceptions import KloppyError


class DatasetTransformer:
    def __init__(
        self,
        from_coordinate_system: CoordinateSystem = None,
        from_pitch_dimensions: PitchDimensions = None,
        from_orientation: Orientation = None,
        to_coordinate_system: CoordinateSystem = None,
        to_pitch_dimensions: PitchDimensions = None,
        to_orientation: Orientation = None,
    ):

        if (
            from_pitch_dimensions
            and from_coordinate_system
            or to_pitch_dimensions
            and to_coordinate_system
        ):
            raise ValueError(
                "You can't specify both a PitchDimension and CoordinateSysetm transformation on the same Transformer instance"
            )

        self._from_coordinate_system = from_coordinate_system
        self._from_pitch_dimensions = (
            from_pitch_dimensions
            if from_pitch_dimensions
            else from_coordinate_system.pitch_dimensions
        )
        self._from_orientation = from_orientation
        self._to_coordinate_system = to_coordinate_system
        self._to_pitch_dimensions = (
            to_pitch_dimensions
            if to_pitch_dimensions
            else to_coordinate_system.pitch_dimensions
        )
        self._to_orientation = to_orientation

    @property
    def _needs_coordinate_system_change(self):
        return self._from_coordinate_system != self._to_coordinate_system

    @property
    def _needs_pitch_dimensions_change(self):
        if self._from_coordinate_system and self._to_coordinate_system:
            return (
                self._from_coordinate_system.pitch_dimensions
                != self._to_coordinate_system.pitch_dimensions
            )

        if self._from_pitch_dimensions and self._to_pitch_dimensions:
            return self._from_pitch_dimensions != self._to_pitch_dimensions

    def change_point_dimensions(self, point: Union[Point, Point3D]) -> Point:

        if point is None:
            return None

        x_base = self._from_pitch_dimensions.x_dim.to_base(point.x)
        y_base = self._from_pitch_dimensions.y_dim.to_base(point.y)

        x = self._to_pitch_dimensions.x_dim.from_base(x_base)
        y = self._to_pitch_dimensions.y_dim.from_base(y_base)

        if isinstance(point, Point3D):
            return Point3D(x=x, y=y, z=point.z)
        else:
            return Point(x=x, y=y)

    def flip_point(self, point: Union[Point, Point3D]):

        if not point:
            return None

        x_base = self._to_pitch_dimensions.x_dim.to_base(point.x)
        y_base = self._to_pitch_dimensions.y_dim.to_base(point.y)

        x_base = 1 - x_base
        y_base = 1 - y_base

        x = self._to_pitch_dimensions.x_dim.from_base(x_base)
        y = self._to_pitch_dimensions.y_dim.from_base(y_base)

        if isinstance(point, Point3D):
            return Point3D(x=x, y=y, z=point.z)
        else:
            return Point(x=x, y=y)

    def __needs_flip(
        self,
        ball_owning_team: Team,
        attacking_direction: AttackingDirection,
        action_executing_team: Team = None,
    ) -> bool:
        if self._from_orientation == self._to_orientation:
            flip = False
        else:
            if not action_executing_team:
                action_executing_team = ball_owning_team

            orientation_factor_from = (
                self._from_orientation.get_orientation_factor(
                    ball_owning_team=ball_owning_team,
                    attacking_direction=attacking_direction,
                    action_executing_team=action_executing_team,
                )
            )
            orientation_factor_to = (
                self._to_orientation.get_orientation_factor(
                    ball_owning_team=ball_owning_team,
                    attacking_direction=attacking_direction,
                    action_executing_team=action_executing_team,
                )
            )
            flip = orientation_factor_from != orientation_factor_to
        return flip

    def transform_frame(self, frame: Frame) -> Frame:

        # Change coordinate system
        if self._needs_coordinate_system_change:
            frame = self.__change_frame_coordinate_system(frame)
        # Change dimensions
        elif self._needs_pitch_dimensions_change:
            frame = self.__change_frame_dimensions(frame)

        # Flip frame based on orientation
        if self.__needs_flip(
            ball_owning_team=frame.ball_owning_team,
            attacking_direction=frame.period.attacking_direction,
        ):
            frame = self.__flip_frame(frame)

        return frame

    def __change_frame_coordinate_system(self, frame: Frame):

        return Frame(
            # doesn't change
            timestamp=frame.timestamp,
            frame_id=frame.frame_id,
            ball_owning_team=frame.ball_owning_team,
            ball_state=frame.ball_state,
            period=frame.period,
            # changes
            ball_coordinates=self.__change_point_coordinate_system(
                frame.ball_coordinates
            ),
            players_data={
                key: PlayerData(
                    coordinates=self.__change_point_coordinate_system(
                        player_data.coordinates
                    ),
                    distance=player_data.distance,
                    speed=player_data.speed,
                    other_data=player_data.other_data,
                )
                for key, player_data in frame.players_data.items()
            },
            other_data=frame.other_data,
        )

    def __change_frame_dimensions(self, frame: Frame):

        return Frame(
            # doesn't change
            timestamp=frame.timestamp,
            frame_id=frame.frame_id,
            ball_owning_team=frame.ball_owning_team,
            ball_state=frame.ball_state,
            period=frame.period,
            # changes
            ball_coordinates=self.change_point_dimensions(
                frame.ball_coordinates
            ),
            players_data={
                key: PlayerData(
                    coordinates=self.change_point_dimensions(
                        player_data.coordinates
                    ),
                    distance=player_data.distance,
                    speed=player_data.speed,
                    other_data=player_data.other_data,
                )
                for key, player_data in frame.players_data.items()
            },
            other_data=frame.other_data,
        )

    def __change_point_coordinate_system(self, point: Union[Point, Point3D]):

        if not point:
            return None

        x = self._from_coordinate_system.pitch_dimensions.x_dim.to_base(
            point.x
        )
        y = self._from_coordinate_system.pitch_dimensions.y_dim.to_base(
            point.y
        )

        if (
            self._from_coordinate_system.vertical_orientation
            != self._to_coordinate_system.vertical_orientation
        ):
            y = 1 - y

        if not self._to_coordinate_system.normalized:
            x = self._to_coordinate_system.pitch_dimensions.x_dim.from_base(x)
            y = self._to_coordinate_system.pitch_dimensions.y_dim.from_base(y)

        if isinstance(point, Point3D):
            return Point3D(x=x, y=y, z=point.z)
        else:
            return Point(x=x, y=y)

    def __flip_frame(self, frame: Frame):

        players_data = {}
        for player, data in frame.players_data.items():
            players_data[player] = PlayerData(
                coordinates=self.flip_point(data.coordinates),
                distance=data.distance,
                speed=data.speed,
                other_data=data.other_data,
            )

        return Frame(
            # doesn't change
            timestamp=frame.timestamp,
            frame_id=frame.frame_id,
            ball_owning_team=frame.ball_owning_team,
            ball_state=frame.ball_state,
            period=frame.period,
            # changes
            ball_coordinates=self.flip_point(frame.ball_coordinates),
            players_data=players_data,
            other_data=frame.other_data,
        )

    def transform_event(self, event: Event) -> Event:

        # Change coordinate system
        if self._needs_coordinate_system_change:
            event = self.__change_event_coordinate_system(event)
        # Change dimensions
        elif self._needs_pitch_dimensions_change:
            event = self.__change_event_dimensions(event)

        # Flip event based on orientation
        if self.__needs_flip(
            ball_owning_team=event.ball_owning_team,
            attacking_direction=event.period.attacking_direction,
            action_executing_team=event.team,
        ):
            event = self.__flip_event(event)

        if event.freeze_frame:
            event.freeze_frame = self.transform_frame(event.freeze_frame)

        return event

    def __change_event_coordinate_system(self, event: Event):

        position_changes = {
            field.name: self.__change_point_coordinate_system(
                getattr(event, field.name)
            )
            for field in fields(event)
            if field.name.endswith("coordinates")
            and getattr(event, field.name)
        }

        return replace(event, **position_changes)

    def __change_event_dimensions(self, event: Event):

        position_changes = {
            field.name: self.change_point_dimensions(
                getattr(event, field.name)
            )
            for field in fields(event)
            if field.name.endswith("coordinates")
            and getattr(event, field.name)
        }

        return replace(event, **position_changes)

    def __flip_event(self, event: Event):

        position_changes = {
            field.name: self.flip_point(getattr(event, field.name))
            for field in fields(event)
            if field.name.endswith("coordinates")
            and getattr(event, field.name)
        }

        return replace(event, **position_changes)

    def get_to_coordinate_system(self) -> CoordinateSystem:
        return self._to_coordinate_system

    @classmethod
    def transform_dataset(
        cls,
        dataset: Dataset,
        to_pitch_dimensions: PitchDimensions = None,
        to_orientation: Orientation = None,
        to_coordinate_system: CoordinateSystem = None,
    ) -> Dataset:

        if to_pitch_dimensions and to_coordinate_system:
            raise ValueError(
                "You can't do both a PitchDimension and CoordinateSysetm on the same dataset transformation"
            )

        if (
            not to_pitch_dimensions
            and not to_orientation
            and not to_coordinate_system
        ):
            return dataset
        elif not to_orientation:
            to_orientation = dataset.metadata.orientation

        if to_orientation == Orientation.BALL_OWNING_TEAM:
            if not dataset.metadata.flags & DatasetFlag.BALL_OWNING_TEAM:
                raise ValueError(
                    "Cannot transform to BALL_OWNING_TEAM orientation when dataset doesn't contain "
                    "ball owning team data"
                )

        if to_pitch_dimensions:

            transformer = cls(
                from_pitch_dimensions=dataset.metadata.pitch_dimensions,
                from_orientation=dataset.metadata.orientation,
                to_pitch_dimensions=to_pitch_dimensions,
                to_orientation=to_orientation,
            )
            metadata = replace(
                dataset.metadata,
                pitch_dimensions=to_pitch_dimensions,
                orientation=to_orientation,
            )

        elif to_coordinate_system:

            transformer = cls(
                from_coordinate_system=dataset.metadata.coordinate_system,
                from_orientation=dataset.metadata.orientation,
                to_coordinate_system=to_coordinate_system,
                to_orientation=to_orientation,
            )
            metadata = replace(
                dataset.metadata,
                coordinate_system=to_coordinate_system,
                pitch_dimensions=to_coordinate_system.pitch_dimensions,
                orientation=to_orientation,
            )

        else:

            transformer = cls(
                from_coordinate_system=dataset.metadata.coordinate_system,
                from_orientation=dataset.metadata.orientation,
                to_coordinate_system=dataset.metadata.coordinate_system,
                to_orientation=to_orientation,
            )
            metadata = replace(
                dataset.metadata,
                orientation=to_orientation,
            )

        if isinstance(dataset, TrackingDataset):
            frames = [
                transformer.transform_frame(record)
                for record in dataset.records
            ]

            return TrackingDataset(
                metadata=metadata,
                records=frames,
            )
        elif isinstance(dataset, EventDataset):
            events = [
                transformer.transform_event(event) for event in dataset.records
            ]

            return EventDataset(
                metadata=metadata,
                records=events,
            )
        else:
            raise KloppyError("Unknown Dataset type")



from dataclasses import dataclass, field
from typing import List, Dict, Callable, Union, Any

from kloppy.domain.models.common import DatasetType

from .common import Dataset, DataRecord
from ...utils import docstring_inherit_attributes


@dataclass
@docstring_inherit_attributes(DataRecord)
class Code(DataRecord):
    """
    Single code

    Attributes:
        id: identifier provided by the coding software
        code: string describing the code
        end_timestamp: float
        labels: Text labels describing this code instance

    """

    code_id: str
    code: str
    end_timestamp: float
    labels: Dict[str, Union[bool, str]] = field(default_factory=dict)

    @property
    def record_id(self) -> str:
        return self.record_id

    @property
    def start_timestamp(self):
        return self.timestamp


@dataclass
class CodeDataset(Dataset[Code]):
    records: List[Code]

    dataset_type: DatasetType = DatasetType.CODE

    @property
    def codes(self):
        return self.records

    def to_pandas(
        self,
        record_converter: Callable[[Code], Dict] = None,
        additional_columns: Dict[
            str, Union[Callable[[Code], Any], Any]
        ] = None,
    ) -> "DataFrame":
        try:
            import pandas as pd
        except ImportError:
            raise ImportError(
                "Seems like you don't have pandas installed. Please"
                " install it using: pip install pandas"
            )

        if not record_converter:

            def record_converter(code: Code) -> Dict:
                row = dict(
                    code_id=code.code_id,
                    period_id=code.period.id if code.period else None,
                    timestamp=code.timestamp,
                    end_timestamp=code.end_timestamp,
                    code=code.code,
                )
                row.update(code.labels)

                return row

        def generic_record_converter(code: Code):
            row = record_converter(code)
            if additional_columns:
                for k, v in additional_columns.items():
                    if callable(v):
                        value = v(code)
                    else:
                        value = v
                    row.update({k: value})

            return row

        return pd.DataFrame.from_records(
            map(generic_record_converter, self.records)
        )


__all__ = ["Code", "CodeDataset"]



import os
from collections import defaultdict

import pytest

from kloppy.domain import (
    AttackingDirection,
    BodyPart,
    BodyPartQualifier,
    DatasetType,
    Orientation,
    Period,
    Point,
    Provider,
    FormationType,
    Frame,
    Position,
)

from kloppy import statsbomb
from kloppy.domain.models.event import (
    CardType,
    PassQualifier,
    PassType,
)


class TestStatsBomb:
    """"""

    @pytest.fixture
    def event_data(self) -> str:
        base_dir = os.path.dirname(__file__)
        return f"{base_dir}/files/statsbomb_event.json"

    @pytest.fixture
    def lineup_data(self) -> str:
        base_dir = os.path.dirname(__file__)
        return f"{base_dir}/files/statsbomb_lineup.json"

    def test_correct_deserialization(self, lineup_data: str, event_data: str):
        """
        This test uses data from the StatsBomb open data project.
        """
        dataset = statsbomb.load(
            lineup_data=lineup_data,
            event_data=event_data,
            coordinates="statsbomb",
        )

        assert dataset.metadata.provider == Provider.STATSBOMB
        assert dataset.dataset_type == DatasetType.EVENT
        assert len(dataset.events) == 4023
        assert len(dataset.metadata.periods) == 2
        assert (
            dataset.metadata.orientation == Orientation.ACTION_EXECUTING_TEAM
        )
        assert dataset.metadata.teams[0].name == "Barcelona"
        assert dataset.metadata.teams[1].name == "Deportivo Alavés"
        assert dataset.metadata.teams[0].starting_formation == FormationType(
            "4-4-2"
        )
        assert dataset.metadata.teams[1].starting_formation == FormationType(
            "4-1-4-1"
        )

        player = dataset.metadata.teams[0].get_player_by_id("5503")
        assert player.player_id == "5503"
        assert player.jersey_no == 10
        assert str(player) == "Lionel Andrés Messi Cuccittini"
        assert player.position == Position(
            position_id="24", name="Left Center Forward", coordinates=None
        )
        assert player.starting

        sub_player = dataset.metadata.teams[0].get_player_by_id("3501")
        assert str(sub_player) == "Philippe Coutinho Correia"
        assert not sub_player.starting

        assert dataset.metadata.periods[0] == Period(
            id=1,
            start_timestamp=0.0,
            end_timestamp=2705.267,
            attacking_direction=AttackingDirection.NOT_SET,
        )
        assert dataset.metadata.periods[1] == Period(
            id=2,
            start_timestamp=2705.268,
            end_timestamp=5557.321,
            attacking_direction=AttackingDirection.NOT_SET,
        )

        assert dataset.events[10].coordinates == Point(34.5, 20.5)

        assert (
            dataset.events[792].get_qualifier_value(BodyPartQualifier)
            == BodyPart.HEAD
        )

        assert (
            dataset.events[2232].get_qualifier_value(BodyPartQualifier)
            == BodyPart.RIGHT_FOOT
        )

        assert (
            dataset.events[195].get_qualifier_value(BodyPartQualifier) is None
        )

        assert (
            dataset.events[1433].get_qualifier_value(PassQualifier)
            == PassType.CROSS
        )

        assert (
            dataset.events[1552].get_qualifier_value(PassQualifier)
            == PassType.THROUGH_BALL
        )

        assert (
            dataset.events[443].get_qualifier_value(PassQualifier)
            == PassType.SWITCH_OF_PLAY
        )

        assert (
            dataset.events[3438].get_qualifier_value(PassQualifier)
            == PassType.LONG_BALL
        )

        assert (
            dataset.events[2266].get_qualifier_value(PassQualifier)
            == PassType.HIGH_PASS
        )

        assert (
            dataset.events[653].get_qualifier_value(PassQualifier)
            == PassType.HEAD_PASS
        )

        assert (
            dataset.events[3134].get_qualifier_value(PassQualifier)
            == PassType.HAND_PASS
        )

        assert (
            dataset.events[3611].get_qualifier_value(PassQualifier)
            == PassType.ASSIST
        )

        assert dataset.events[3392].get_qualifier_value(PassQualifier) is None

    def test_correct_normalized_deserialization(
        self, lineup_data: str, event_data: str
    ):
        """
        This test uses data from the StatsBomb open data project.
        """
        dataset = statsbomb.load(
            lineup_data=lineup_data, event_data=event_data
        )

        assert dataset.events[10].coordinates == Point(0.2875, 0.25625)

    def test_substitution(self, lineup_data: str, event_data: str):
        """
        Test substitution events
        """
        dataset = statsbomb.load(
            lineup_data=lineup_data,
            event_data=event_data,
            event_types=["substitution"],
        )

        assert len(dataset.events) == 6

        subs = [
            (6374, 3501),
            (6839, 6935),
            (6581, 6566),
            (6613, 6624),
            (5477, 11392),
            (5203, 8206),
        ]

        for event_idx, (player_id, replacement_player_id) in enumerate(subs):
            event = dataset.events[event_idx]
            assert event.player == event.team.get_player_by_id(player_id)
            assert event.replacement_player == event.team.get_player_by_id(
                replacement_player_id
            )

    def test_card(self, lineup_data: str, event_data: str):
        """
        Test card events
        """
        dataset = statsbomb.load(
            lineup_data=lineup_data,
            event_data=event_data,
            event_types=["card"],
        )

        assert len(dataset.events) == 2

        for card in dataset.events:
            assert card.card_type == CardType.FIRST_YELLOW

    def test_foul_committed(self, lineup_data: str, event_data: str):
        """
        Test foul committed events
        """
        dataset = statsbomb.load(
            lineup_data=lineup_data,
            event_data=event_data,
            event_types=["foul_committed"],
        )

        assert len(dataset.events) == 23

    def test_related_events(self, lineup_data: str, event_data: str):
        dataset = statsbomb.load(
            lineup_data=lineup_data, event_data=event_data
        )
        carry_event = dataset.get_event_by_id(
            "8e3dacc2-7a39-4301-9053-e78cfec1aa95"
        )
        pass_event = dataset.get_event_by_id(
            "d1cccb73-c7ef-4b02-8267-ebd7f149904b"
        )
        receipt_event = dataset.get_event_by_id(
            "61da36dc-d862-416c-8ee3-1a0cd24dc086"
        )

        assert carry_event.get_related_events() == [receipt_event, pass_event]
        assert carry_event.related_pass() == pass_event

    def test_player_position(self, lineup_data: str, event_data: str):
        """
        Validate position information is loaded correctly from the STARTING_XI events.

        TODO: Fix when player is substituted
        """
        dataset = statsbomb.load(
            lineup_data=lineup_data,
            event_data=event_data,
        )
        player_5246 = dataset.metadata.teams[0].get_player_by_id(20055)
        assert player_5246.position.position_id == "1"

    def test_freeze_frame(self, lineup_data: str, event_data: str):
        """
        Test freeze-frame is properly loaded and attached to shot events. The
        freeze-frames contain location information on player level.
        """
        dataset = statsbomb.load(
            lineup_data=lineup_data,
            event_data=event_data,
            coordinates="statsbomb",
        )
        shot_event = dataset.get_event_by_id(
            "65f16e50-7c5d-4293-b2fc-d20887a772f9"
        )

        event_player_coordinates = shot_event.freeze_frame.players_coordinates[
            shot_event.player
        ]
        assert event_player_coordinates == shot_event.coordinates

        player_5246 = dataset.metadata.teams[0].get_player_by_id(5246)
        assert shot_event.freeze_frame.players_coordinates[
            player_5246
        ] == Point(103.2, 43.6)

        assert shot_event.freeze_frame.frame_id == 3727

    def test_freeze_frame_360(self):
        base_dir = os.path.dirname(__file__)

        dataset = statsbomb.load(
            event_data=f"{base_dir}/files/statsbomb_3788741_event.json",
            lineup_data=f"{base_dir}/files/statsbomb_3788741_lineup.json",
            three_sixty_data=f"{base_dir}/files/statsbomb_3788741_360.json",
            coordinates="statsbomb",
        )

        pass_event = dataset.find("pass")
        coordinates_per_team = defaultdict(list)
        for (
            player,
            coordinates,
        ) in pass_event.freeze_frame.players_coordinates.items():
            coordinates_per_team[player.team.name].append(coordinates)

        assert coordinates_per_team == {
            "Italy": [
                Point(x=43.672042000000005, y=31.609489),
                Point(x=44.016185, y=45.22054),
                Point(x=49.69454, y=35.055324000000006),
                Point(x=54.51254, y=29.714278),
                Point(x=58.29811, y=48.149497000000004),
                Point(x=60.362968, y=30.403444999999998),
            ],
            "Turkey": [
                Point(x=60.018825, y=39.621055000000005),
                Point(x=61.223323, y=47.977206),
                Point(x=67.417896, y=32.643237000000006),
                Point(x=68.96654000000001, y=43.325328000000006),
            ],
        }

        assert pass_event.freeze_frame.players_coordinates[
            pass_event.player
        ] == Point(x=60.018825, y=39.621055000000005)

        assert pass_event.freeze_frame.frame_id == 21



from typing import List
from lxml import etree

from cssselect import GenericTranslator

from kloppy.domain import Event, EventType


class CSSPatternMatcher:
    def __init__(self, pattern: str):
        self.expression = GenericTranslator().css_to_xpath([pattern])

    def match(self, events: List[Event]) -> List[List[Event]]:
        elm = etree.Element("start")
        root = elm
        for i, event in enumerate(events):
            if event.event_type != EventType.GENERIC:
                elm = etree.SubElement(
                    elm,
                    event.event_name.lower()
                    .replace(" ", "_")
                    .replace("*", ""),
                    index=i,
                    result=str(event.result).lower(),
                    team=str(event.team.ground).lower(),
                    attrib={
                        "class": str(event.result).lower()
                        + " "
                        + str(event.team.ground).lower()
                    },
                )

        matching_events = []
        for elm in root.xpath(self.expression):
            matching_events.append(events[elm.attrib["index"]])
        return matching_events



import os
import re

import pytest
from pandas import DataFrame
from lxml import objectify

from kloppy.domain import (
    Point,
    Point3D,
    Provider,
)
from kloppy.infra.serializers.tracking.metrica_epts.metadata import (
    load_metadata,
)
from kloppy.infra.serializers.tracking.metrica_epts.metadata import (
    _load_provider,
)
from kloppy.infra.serializers.tracking.metrica_epts.reader import (
    build_regex,
    read_raw_data,
)
from kloppy.utils import performance_logging


from kloppy import metrica


class TestMetricaEPTSTracking:
    def test_regex(self):
        base_dir = os.path.dirname(__file__)
        with open(
            f"{base_dir}/files/epts_metrica_metadata.xml", "rb"
        ) as metadata_fp:
            metadata = load_metadata(metadata_fp)

        regex_str = build_regex(
            metadata.data_format_specifications[0],
            metadata.player_channels,
            metadata.sensors,
        )

        regex = re.compile(regex_str)

        result = regex.search(
            "450:0.30602,0.97029,5,1.1127902269363403;0.48496,0.73308,1,5.2345757484436035;0.38851,0.97786,2,3.32438325881958;0.48414,0.92730,0,5.878201961517334;0.25661,0.60984,8,4.325275421142578;0.24170,0.49115,7,1.9475973844528198;0.38878,0.47540,4,1.8947187662124634;0.31854,0.89154,3,2.741175413131714;NaN,NaN,-1,NaN;0.25658,0.73095,9,3.418901205062866;NaN,NaN,10,NaN;0.32513,0.89532,19,0.6292267441749573;0.30286,0.98231,17,1.287971019744873;0.39954,0.93079,12,3.8761117458343506;0.23961,0.63742,20,1.6785365343093872;0.38319,0.81908,14,3.611333131790161;0.49984,0.91227,13,5.139825820922852;0.46733,0.49708,16,3.1186790466308594;0.26311,0.48218,18,2.309934139251709;0.53767,0.81373,11,6.402815818786621;0.39631,0.77277,15,3.338606357574463;NaN,NaN,21,NaN:0.52867,0.70690,NaN"
        )

        assert result is not None

    def test_provider_name_recognition(self):
        base_dir = os.path.dirname(__file__)
        with open(
            f"{base_dir}/files/epts_metrica_metadata.xml", "rb"
        ) as metadata_fp:
            root = objectify.fromstring(metadata_fp.read())
            metadata = root.find("Metadata")
            provider_from_file = _load_provider(metadata)

        assert provider_from_file == Provider.METRICA

    def test_read(self):
        base_dir = os.path.dirname(__file__)
        with open(
            f"{base_dir}/files/epts_metrica_metadata.xml", "rb"
        ) as metadata_fp:
            metadata = load_metadata(metadata_fp)

        with open(
            f"{base_dir}/files/epts_metrica_tracking.txt", "rb"
        ) as raw_data:
            iterator = read_raw_data(raw_data, metadata)

            with performance_logging("load"):
                assert list(iterator)

    def test_read_to_pandas(self):
        base_dir = os.path.dirname(__file__)

        with open(
            f"{base_dir}/files/epts_metrica_metadata.xml", "rb"
        ) as metadata_fp, open(
            f"{base_dir}/files/epts_metrica_tracking.txt", "rb"
        ) as raw_data:

            metadata = load_metadata(metadata_fp)
            records = read_raw_data(
                raw_data, metadata, sensor_ids=["position"]
            )
            data_frame = DataFrame.from_records(records)

        assert "player_Track_1_x" in data_frame.columns

    def test_skip_sensors(self):
        base_dir = os.path.dirname(__file__)

        with open(
            f"{base_dir}/files/epts_metrica_metadata.xml", "rb"
        ) as metadata_fp, open(
            f"{base_dir}/files/epts_metrica_tracking.txt", "rb"
        ) as raw_data:
            metadata = load_metadata(metadata_fp)
            records = read_raw_data(raw_data, metadata, sensor_ids=["speed"])
            data_frame = DataFrame.from_records(records)

        assert "player_Track_1_s" in data_frame.columns
        assert "player_Track_1_x" not in data_frame.columns

    @pytest.fixture
    def meta_data(self) -> str:
        base_dir = os.path.dirname(__file__)
        return f"{base_dir}/files/epts_metrica_metadata.xml"

    @pytest.fixture
    def raw_data(self) -> str:
        base_dir = os.path.dirname(__file__)
        return f"{base_dir}/files/epts_metrica_tracking.txt"

    def test_correct_deserialization(self, meta_data: str, raw_data: str):
        dataset = metrica.load_tracking_epts(
            meta_data=meta_data, raw_data=raw_data
        )

        first_player = next(iter(dataset.records[0].players_data))

        assert len(dataset.records) == 100
        assert len(dataset.metadata.periods) == 2
        assert dataset.metadata.orientation is None

        assert dataset.records[0].players_data[
            first_player
        ].coordinates == Point(x=0.30602, y=0.97029)

        assert dataset.records[0].ball_coordinates == Point3D(
            x=0.52867, y=0.7069, z=None
        )

    def test_other_data_deserialization(self, meta_data: str, raw_data: str):
        dataset = metrica.load_tracking_epts(
            meta_data=meta_data, raw_data=raw_data
        )

        first_player = next(iter(dataset.records[0].players_data))

        assert (
            dataset.records[0].players_data[first_player].other_data["mapping"]
            == 5.0
        )

    def test_read_with_sensor_unused_in_players_and_frame_count_name_modified(
        self,
    ):
        base_dir = os.path.dirname(__file__)

        with open(
            f"{base_dir}/files/epts_metrica_metadata_unused_sensor.xml", "rb"
        ) as metadata_fp, open(
            f"{base_dir}/files/epts_metrica_tracking.txt", "rb"
        ) as raw_data:
            dataset = metrica.load_tracking_epts(
                meta_data=metadata_fp, raw_data=raw_data
            )
        # Acceleration field is in other data
        other_data = list(dataset.frames[0].players_data.items())[0][
            1
        ].other_data
        assert "acceleration" in other_data
        # But is None due to there is not channel for this sensor
        assert other_data["acceleration"] is None
        # But all defined sensors in the metadata are 4
        assert len(dataset.metadata.sensors) == 4

    def test_read_empty_player_values(
        self,
    ):
        base_dir = os.path.dirname(__file__)

        with open(
            f"{base_dir}/files/epts_metrica_metadata.xml", "rb"
        ) as metadata_fp, open(
            f"{base_dir}/files/epts_metrica_tracking_with_empty_values.txt",
            "rb",
        ) as raw_data:
            dataset = metrica.load_tracking_epts(
                meta_data=metadata_fp, raw_data=raw_data
            )

        first_player = next(iter(dataset.records[0].players_data))

        first_player_x = (
            dataset.records[0].players_data[first_player].coordinates.x
        )

        first_player_y = (
            dataset.records[0].players_data[first_player].coordinates.y
        )

        # x,y coordinates of the first player are empty. Check if they are nan's
        assert first_player_x != first_player_x
        assert first_player_y != first_player_y
